@@ -1,6 +1,6 @@
 /* Yash: yet another shell */
 /* job.c: job control */
-/* (C) 2007-2024 magicant */
+/* (C) 2007-2025 magicant */
 
 /* This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -114,8 +114,9 @@ void set_active_job(job_T *job)
 /* Moves the active job into the job list.
  * If the newly added job is stopped, it becomes the current job.
  * If `current' is true or there is no current job, the newly added job becomes
- * the current job if there is no stopped job. */
-void add_job(bool current)
+ * the current job if there is no stopped job.
+ * Returns the job number of the added job. */
+size_t add_job(bool current)
 {
     job_T *job = joblist.contents[ACTIVE_JOBNO];
     size_t jobnumber;
@@ -140,6 +141,8 @@ set_current:
         set_current_jobnumber(jobnumber);
     else
         set_current_jobnumber(current_jobnumber);
+
+    return jobnumber;
 }
 
 /* Returns the job of the specified number or NULL if not found. */
@@ -416,22 +419,21 @@ found:
 #endif
 
     /* decide the job status from the process status:
-     * - JS_RUNNING if any of the processes is running.
-     * - JS_STOPPED if no processes are running but some are stopped.
+     * - JS_RUNNING if no processes are stopped but some are running.
+     * - JS_STOPPED if any of the processes is stopped.
      * - JS_DONE if all the processes are finished. */
-    jobstatus_T oldstatus = job->j_status;
-    bool anyrunning = false, anystopped = false;
+    jobstatus_T oldstatus = job->j_status, newstatus = JS_DONE;
     /* check if there are running/stopped processes */
     for (size_t i = 0; i < job->j_pcount; i++) {
         switch (job->j_procs[i].pr_status) {
-            case JS_RUNNING:  anyrunning = true;  goto out_of_loop;
-            case JS_STOPPED:  anystopped = true;  break;
-            default:                              break;
+            case JS_RUNNING:  newstatus = JS_RUNNING;  break;
+            case JS_STOPPED:  newstatus = JS_STOPPED;  goto out_of_loop;
+            case JS_DONE:                              break;
         }
     }
 out_of_loop:
-    job->j_status = anyrunning ? JS_RUNNING : anystopped ? JS_STOPPED : JS_DONE;
-    if (job->j_status != oldstatus)
+    job->j_status = newstatus;
+    if (newstatus != oldstatus)
         job->j_statuschanged = true;
 
     goto start;
@@ -441,6 +443,8 @@ out_of_loop:
  * `jobnumber' must be a valid job number.
  * If `return_on_stop' is false, waits for the job to finish.
  * Otherwise, waits for the job to finish or stop.
+ * When the job is stopped, the E_CANCEL exception is raised if the shell is
+ * interactive.
  * If `interruptible' is true, this function can be canceled by SIGINT.
  * If `return_on_trap' is true, this function returns false immediately after
  * trap actions are performed. Otherwise, traps are not handled.
@@ -462,8 +466,11 @@ int wait_for_job(size_t jobnumber, bool return_on_stop,
         for (;;) {
             if (job->j_status == JS_DONE)
                 break;
-            if (return_on_stop && job->j_status == JS_STOPPED)
+            if (return_on_stop && job->j_status == JS_STOPPED) {
+                if (is_interactive_now)
+                    cancel_current_command();
                 break;
+            }
             signum = wait_for_sigchld(interruptible, return_on_trap);
             if (signum != 0)
                 break;
@@ -486,6 +493,7 @@ int wait_for_job(size_t jobnumber, bool return_on_stop,
  * a pointer to a pointer to a wide string. The caller must assign a pointer to
  * a newly malloced wide string to the variable the return value points to.
  * This string is used as the name of the new stopped job.
+ * Additionally, if the shell is interactive, the E_CANCEL exception is raised.
  * If the child exited, this function returns NULL.
  * The exit status is assigned to `laststatus' in any case. */
 wchar_t **wait_for_child(pid_t cpid, pid_t cpgid, bool return_on_stop)
@@ -561,38 +569,64 @@ void put_foreground(pid_t pgrp)
     sigprocmask(SIG_SETMASK, &savess, NULL);
 }
 
-/* Ensures the current shell process is in the foreground.
- * The shell process is stopped by SIGTTOU until it is put in the foreground.
- * This function requires `doing_job_control_now' to be true. */
-/* This function prevents the job-control shell from mangling the terminal while
- * another shell is using it. */
+/* Ensures that the process group of the current shell process is the foreground
+ * process group of the controlling terminal.
+ * This function requires `doing_job_control_now' to be true.
+ * This function calls `tcsetpgrp' to bring the current process group to the
+ * foreground. If the process group is the session leader's group, this is done
+ * while the shell is blocking SIGTTOU to prevent the shell from being stopped.
+ * If the process group is not the session leader's group, the call is made
+ * without blocking or ignoring SIGTTOU, so the shell will be stopped if it is
+ * in the background. In this case, the shell has to be brought back to the
+ * foreground by some other process. In either case, the shell is in the
+ * foreground when this function returns. */
+/* This function implements part of the initialization of the job-control shell.
+ * POSIX says that the shell should bring itself into the foreground only if it
+ * is started as the controlling process (that is, the session leader) for the
+ * terminal session. However, this function also brings the shell into the
+ * foreground if the shell is in the same process group as the session leader
+ * because it is unlikely that there is another job-controlling process that can
+ * bring the shell into the foreground. */
+/* POSIX also requires that the shell use SIGTTIN to suspend itself when it is
+ * in the background. However, we use SIGTTOU instead for simplicity and
+ * stability. Using `tcsetpgrp' avoids the TOCTOU issue between the check for
+ * the foreground-ness of the shell and the sending of the stop signal. */
 void ensure_foreground(void)
 {
-    /* This function calls `tcsetpgrp' with the default SIGTTOU handler. If the
-     * shell is in the background, it will receive SIGTTOU and get stopped until
-     * it is continued in the foreground. */
-
     struct sigaction dflsa, savesa;
     sigset_t blockss, savess;
 
     assert(doing_job_control_now);
     assert(shell_pgid > 0);
 
-    dflsa.sa_handler = SIG_DFL;
-    dflsa.sa_flags = 0;
-    sigemptyset(&dflsa.sa_mask);
-    sigemptyset(&savesa.sa_mask);
-    sigaction(SIGTTOU, &dflsa, &savesa);
+    if (shell_pgid == getsid(0)) {
+        /* The shell is in the session leader's process group. */
+        sigemptyset(&blockss);
+        sigaddset(&blockss, SIGTTOU);
+        sigemptyset(&savess);
+        sigprocmask(SIG_BLOCK, &blockss, &savess);
 
-    sigemptyset(&blockss);
-    sigaddset(&blockss, SIGTTOU);
-    sigemptyset(&savess);
-    sigprocmask(SIG_UNBLOCK, &blockss, &savess);
+        tcsetpgrp(ttyfd, shell_pgid);
 
-    tcsetpgrp(ttyfd, shell_pgid);
+        sigprocmask(SIG_SETMASK, &savess, NULL);
+    } else {
+        /* The shell is not in the session leader's process group. */
+        dflsa.sa_handler = SIG_DFL;
+        dflsa.sa_flags = 0;
+        sigemptyset(&dflsa.sa_mask);
+        sigemptyset(&savesa.sa_mask);
+        sigaction(SIGTTOU, &dflsa, &savesa);
 
-    sigprocmask(SIG_SETMASK, &savess, NULL);
-    sigaction(SIGTTOU, &savesa, NULL);
+        sigemptyset(&blockss);
+        sigaddset(&blockss, SIGTTOU);
+        sigemptyset(&savess);
+        sigprocmask(SIG_UNBLOCK, &blockss, &savess);
+
+        tcsetpgrp(ttyfd, shell_pgid);
+
+        sigprocmask(SIG_SETMASK, &savess, NULL);
+        sigaction(SIGTTOU, &savesa, NULL);
+    }
 }
 
 /* Computes the exit status from the status code returned by `waitpid'. */
@@ -1345,20 +1379,29 @@ int wait_for_job_by_jobspec(const wchar_t *jobspec)
             || job->j_legacy)
         return Exit_NOTFOUND;
 
-    int signal = wait_for_job(jobnumber,
-            doing_job_control_now, doing_job_control_now, true);
-    if (signal != 0) {
-        assert(TERMSIGOFFSET >= 128);
-        return -(signal + TERMSIGOFFSET);
+    while (job->j_status != JS_DONE) {
+        bool savenonotify = job->j_nonotify;
+        job->j_nonotify = true;
+        int signal = wait_for_sigchld(
+                /* interruptible */ doing_job_control_now,
+                /* return_on_stop */ true);
+        job->j_nonotify = savenonotify;
+        if (signal != 0) {
+            assert(TERMSIGOFFSET >= 128);
+            return -(signal + TERMSIGOFFSET);
+        }
+
+        if (doing_job_control_now && is_interactive_now && !posixly_correct)
+            print_job_status(
+                    jobnumber,
+                    /* changedonly */ true,
+                    /* verbose */ false,
+                    /* remove_done */ false,
+                    stdout);
     }
 
     int status = calc_status_of_job(job);
-    if (job->j_status != JS_RUNNING) {
-        if (doing_job_control_now && is_interactive_now && !posixly_correct)
-            print_job_status(jobnumber, false, false, true, stdout);
-        else if (job->j_status == JS_DONE)
-            remove_job(jobnumber);
-    }
+    remove_job(jobnumber);
     return status;
 }
 
@@ -1369,7 +1412,12 @@ bool wait_builtin_has_job(bool jobcontrol)
     for (size_t i = 1; i < joblist.length; i++) {
         job_T *job = joblist.contents[i];
         if (jobcontrol && is_interactive_now && !posixly_correct)
-            print_job_status(i, true, false, false, stdout);
+            print_job_status(
+                    i,
+                    /* changedonly */ true,
+                    /* verbose */ false,
+                    /* remove_done */ false,
+                    stdout);
         if (job != NULL && (job->j_legacy || job->j_status == JS_DONE))
             remove_job(i);
     }
@@ -1377,7 +1425,7 @@ bool wait_builtin_has_job(bool jobcontrol)
     /* see if we have jobs to wait for. */
     for (size_t i = 1; i < joblist.length; i++) {
         job_T *job = joblist.contents[i];
-        if (job != NULL && (!jobcontrol || job->j_status == JS_RUNNING))
+        if (job != NULL)
             return true;
     }
     return false;
