@@ -1,6 +1,6 @@
 /* Yash: yet another shell */
 /* editing.h: main editing module */
-/* (C) 2007-2016 magicant */
+/* (C) 2007-2017 magicant */
 
 /* This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -150,6 +150,9 @@ extern le_command_func_T
     cmd_complete_list, /*C*/
     cmd_complete_all, /*C*/
     cmd_complete_max, /*C*/
+    cmd_complete_max_then_list, /*C*/
+    cmd_complete_max_then_next_candidate, /*C*/
+    cmd_complete_max_then_prev_candidate, /*C*/
     cmd_clear_candidates; /*C*/
 
 /* vi-mode specific commands */
@@ -191,7 +194,9 @@ extern le_command_func_T
     cmd_emacs_delete_horizontal_space, /*C*/
     cmd_emacs_just_one_space, /*C*/
     cmd_emacs_search_forward, /*C*/
-    cmd_emacs_search_backward; /*C*/
+    cmd_emacs_search_backward, /*C*/
+    cmd_emacs_search_forward_current, /*C*/
+    cmd_emacs_search_backward_current; /*C*/
 
 /* history-related commands */
 extern le_command_func_T
@@ -229,4 +234,4 @@ extern le_command_func_T
 #endif /* YASH_EDITING_H */
 
 
-/* vim: set ts=8 sts=4 sw=4 noet tw=80: */
+/* vim: set ts=8 sts=4 sw=4 et tw=80: */
